
# THIS FILE IS GENERATED FROM SKNANO SETUP.PY
short_version = '0.2.16'
version = '0.2.16'
full_version = '0.2.16'
git_revision = '320b954a5fafd0105e7a0ae22efa7df5202f48fa'
release = True

if not release:
    version = full_version
